from __future__ import annotations

import asyncio
import logging
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Final,
    Iterable,
    Mapping,
    Optional,
    Set,
    Tuple,
)

import aiohttp_cors
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web
from aiohttp_sse import sse_response
from aiotools import adefer
from sqlalchemy.orm import load_only

from ai.backend.common import validators as tx
from ai.backend.common.events.dispatcher import (
    EventDispatcher,
)
from ai.backend.common.events.event_types.kernel.broadcast import (
    KernelCancelledBroadcastEvent,
    KernelCreatingBroadcastEvent,
    KernelPreparingBroadcastEvent,
    KernelPullingBroadcastEvent,
    KernelStartedBroadcastEvent,
    KernelTerminatedBroadcastEvent,
    KernelTerminatingBroadcastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SessionCancelledBroadcastEvent,
    SessionEnqueuedBroadcastEvent,
    SessionFailureBroadcastEvent,
    SessionScheduledBroadcastEvent,
    SessionStartedBroadcastEvent,
    SessionSuccessBroadcastEvent,
    SessionTerminatedBroadcastEvent,
    SessionTerminatingBroadcastEvent,
)
from ai.backend.common.events.hub.propagators.cache import WithCachePropagator
from ai.backend.common.events.types import EventCacheDomain, EventDomain
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter

from ..errors.exceptions import GenericForbidden, GroupNotFound
from ..models import UserRole, groups, kernels
from ..models.session import SessionRow
from ..models.utils import execute_with_retry
from ..types import Sentinel
from .auth import auth_required
from .manager import READ_ALLOWED, server_status_required
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext
    from .types import CORSOptions, WebMiddleware

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

sentinel: Final = Sentinel.token

SessionEventInfo = Tuple[str, dict, str, Optional[int]]


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["name", "sessionName"], default="*") >> "session_name": t.String,
        t.Key("ownerAccessKey", default=None) >> "owner_access_key": t.Null | t.String,
        t.Key("sessionId", default=None) >> "session_id": t.Null | tx.UUID,
        # NOTE: if set, sessionId overrides sessionName and ownerAccessKey parameters.
        tx.AliasedKey(["group", "groupName"], default="*") >> "group_name": t.String,
        t.Key("scope", default="*"): t.Enum("*", "session", "kernel"),
    })
)
@adefer
async def push_session_events(
    defer,
    request: web.Request,
    params: Mapping[str, Any],
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["events.context"]
    session_name = params["session_name"]
    session_id = params["session_id"]
    scope = params["scope"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    access_key = params["owner_access_key"]
    if access_key is None:
        access_key = request["keypair"]["access_key"]
    if user_role == UserRole.USER:
        if access_key != request["keypair"]["access_key"]:
            raise GenericForbidden
    group_name = params["group_name"]
    my_queue: asyncio.Queue[Sentinel | SessionEventInfo] = asyncio.Queue()
    log.info("PUSH_SESSION_EVENTS (ak:{}, s:{}, g:{})", access_key, session_name, group_name)
    if group_name == "*":
        group_id = "*"
    else:
        async with root_ctx.db.begin_readonly() as conn:
            query = sa.select([groups.c.id]).select_from(groups).where(groups.c.name == group_name)
            result = await conn.execute(query)
            row = result.first()
            if row is None:
                raise GroupNotFound
            group_id = row["id"]
    app_ctx.session_event_queues.add(my_queue)
    defer(lambda: app_ctx.session_event_queues.remove(my_queue))
    async with sse_response(request) as resp:
        try:
            while True:
                evdata = await my_queue.get()
                try:
                    if evdata is sentinel:
                        break
                    event_name, row, reason, exit_code = evdata
                    if user_role in (UserRole.USER, UserRole.ADMIN):
                        if row["domain_name"] != request["user"]["domain_name"]:
                            continue
                    if user_role == UserRole.USER:
                        if row["user_uuid"] != user_uuid:
                            continue
                    if group_id != "*" and row["group_id"] != group_id:
                        continue
                    if scope == "session" and not event_name.startswith("session_"):
                        continue
                    if scope == "kernel" and not event_name.startswith("kernel_"):
                        continue
                    if session_id is not None:
                        if row["session_id"] != session_id:
                            continue
                    else:
                        if session_name != "*" and not (
                            (row["session_name"] == session_name)
                            and (row["access_key"] == access_key)
                        ):
                            continue
                    response_data = {
                        "reason": reason,
                        "sessionName": row["session_name"],
                        "ownerAccessKey": row["access_key"],
                        "sessionId": str(row["session_id"]),
                        "exitCode": exit_code,
                    }
                    if kernel_id := row.get("id"):
                        response_data["kernelId"] = str(kernel_id)
                    if cluster_role := row.get("cluster_role"):
                        response_data["clusterRole"] = cluster_role
                    if cluster_idx := row.get("cluster_idx"):
                        response_data["clusterIdx"] = cluster_idx
                    await resp.send(dump_json_str(response_data), event=event_name)
                finally:
                    my_queue.task_done()
        finally:
            return resp


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["task_id", "taskId"]): tx.UUID,
    })
)
async def push_background_task_events(
    request: web.Request,
    params: Mapping[str, Any],
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    task_id: uuid.UUID = params["task_id"]
    access_key = request["keypair"]["access_key"]
    log.info("PUSH_BACKGROUND_TASK_EVENTS (ak:{}, t:{})", access_key, task_id)
    async with sse_response(request) as resp:
        propagator = WithCachePropagator(root_ctx.event_fetcher)
        root_ctx.event_hub.register_event_propagator(
            propagator, [(EventDomain.BGTASK, str(task_id))]
        )
        try:
            cache_id = EventCacheDomain.BGTASK.cache_id(task_id)
            async for event in propagator.receive(cache_id):
                user_event = event.user_event()
                if user_event is None:
                    log.warning(
                        "Received unsupported user event: {}",
                        event.event_name(),
                    )
                    continue
                await resp.send(
                    dump_json_str(user_event.user_event_mapping()),
                    event=user_event.event_name(),
                    retry=user_event.retry_count(),
                )
                if user_event.is_close_event():
                    log.debug(
                        "Received close event: {}",
                        user_event.event_name(),
                    )
                    break
            await resp.send(dump_json_str({}), event="server_close")
        finally:
            root_ctx.event_hub.unregister_event_propagator(propagator.id())
    return resp


async def enqueue_kernel_creation_status_update(
    app: web.Application,
    source: AgentId,
    event: KernelPreparingBroadcastEvent
    | KernelPullingBroadcastEvent
    | KernelCreatingBroadcastEvent
    | KernelStartedBroadcastEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select([
                    kernels.c.id,
                    kernels.c.session_id,
                    kernels.c.session_name,
                    kernels.c.access_key,
                    kernels.c.cluster_role,
                    kernels.c.cluster_idx,
                    kernels.c.domain_name,
                    kernels.c.group_id,
                    kernels.c.user_uuid,
                ])
                .select_from(kernels)
                .where(
                    (kernels.c.id == event.kernel_id),
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.event_name(), row._mapping, event.reason, None))


async def enqueue_kernel_termination_status_update(
    app: web.Application,
    agent_id: AgentId,
    event: KernelCancelledBroadcastEvent
    | KernelTerminatingBroadcastEvent
    | KernelTerminatedBroadcastEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select([
                    kernels.c.id,
                    kernels.c.session_id,
                    kernels.c.session_name,
                    kernels.c.access_key,
                    kernels.c.cluster_role,
                    kernels.c.cluster_idx,
                    kernels.c.domain_name,
                    kernels.c.group_id,
                    kernels.c.user_uuid,
                ])
                .select_from(kernels)
                .where(
                    (kernels.c.id == event.kernel_id),
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        exit_code = (
            event.exit_code
            if isinstance(event, (KernelTerminatingBroadcastEvent, KernelTerminatedBroadcastEvent))
            else None
        )
        q.put_nowait((event.event_name(), row._mapping, event.reason, exit_code))


async def enqueue_session_creation_status_update(
    app: web.Application,
    source: AgentId,
    event: (
        SessionEnqueuedBroadcastEvent
        | SessionScheduledBroadcastEvent
        | SessionStartedBroadcastEvent
        | SessionCancelledBroadcastEvent
    ),
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch() -> SessionRow | None:
        async with root_ctx.db.begin_readonly_session() as db_session:
            query = (
                sa.select(SessionRow)
                .where(SessionRow.id == event.session_id)
                .options(
                    load_only(
                        SessionRow.id,
                        SessionRow.name,
                        SessionRow.access_key,
                        SessionRow.domain_name,
                        SessionRow.group_id,
                        SessionRow.user_uuid,
                    )
                )
            )
            return await db_session.scalar(query)

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    row_map = {
        "session_id": row.id,
        "session_name": row.name,
        "domain_name": row.domain_name,
        "user_uuid": row.user_uuid,
        "group_id": row.group_id,
        "access_key": row.access_key,
    }
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.event_name(), row_map, event.reason, None))


async def enqueue_session_termination_status_update(
    app: web.Application,
    agent_id: AgentId,
    event: SessionTerminatingBroadcastEvent | SessionTerminatedBroadcastEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch() -> SessionRow | None:
        async with root_ctx.db.begin_readonly_session() as db_session:
            query = (
                sa.select(SessionRow)
                .where(SessionRow.id == event.session_id)
                .options(
                    load_only(
                        SessionRow.id,
                        SessionRow.name,
                        SessionRow.access_key,
                        SessionRow.domain_name,
                        SessionRow.group_id,
                        SessionRow.user_uuid,
                    )
                )
            )
            return await db_session.scalar(query)

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    row_map = {
        "session_id": row.id,
        "session_name": row.name,
        "domain_name": row.domain_name,
        "user_uuid": row.user_uuid,
        "group_id": row.group_id,
        "access_key": row.access_key,
    }
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.event_name(), row_map, event.reason, None))


async def enqueue_batch_task_result_update(
    app: web.Application,
    agent_id: AgentId,
    event: SessionSuccessBroadcastEvent | SessionFailureBroadcastEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch() -> SessionRow | None:
        async with root_ctx.db.begin_readonly_session() as db_session:
            query = (
                sa.select(SessionRow)
                .where(SessionRow.id == event.session_id)
                .options(
                    load_only(
                        SessionRow.id,
                        SessionRow.name,
                        SessionRow.access_key,
                        SessionRow.domain_name,
                        SessionRow.group_id,
                        SessionRow.user_uuid,
                    )
                )
            )
            return await db_session.scalar(query)

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    row_map = {
        "session_id": row.id,
        "session_name": row.name,
        "domain_name": row.domain_name,
        "user_uuid": row.user_uuid,
        "group_id": row.group_id,
        "access_key": row.access_key,
    }
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.event_name(), row_map, event.reason, event.exit_code))


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    session_event_queues: Set[asyncio.Queue[Sentinel | SessionEventInfo]]


async def events_app_ctx(app: web.Application) -> AsyncIterator[None]:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]
    app_ctx.session_event_queues = set()
    event_dispatcher: EventDispatcher = root_ctx.event_dispatcher
    event_dispatcher.subscribe(
        SessionEnqueuedBroadcastEvent, app, enqueue_session_creation_status_update
    )
    event_dispatcher.subscribe(
        SessionScheduledBroadcastEvent, app, enqueue_session_creation_status_update
    )
    event_dispatcher.subscribe(
        KernelPreparingBroadcastEvent, app, enqueue_kernel_creation_status_update
    )
    event_dispatcher.subscribe(
        KernelPullingBroadcastEvent, app, enqueue_kernel_creation_status_update
    )
    event_dispatcher.subscribe(
        KernelCreatingBroadcastEvent, app, enqueue_kernel_creation_status_update
    )
    event_dispatcher.subscribe(
        KernelStartedBroadcastEvent, app, enqueue_kernel_creation_status_update
    )
    event_dispatcher.subscribe(
        SessionStartedBroadcastEvent, app, enqueue_session_creation_status_update
    )
    event_dispatcher.subscribe(
        KernelTerminatingBroadcastEvent, app, enqueue_kernel_termination_status_update
    )
    event_dispatcher.subscribe(
        KernelTerminatedBroadcastEvent, app, enqueue_kernel_termination_status_update
    )
    event_dispatcher.subscribe(
        KernelCancelledBroadcastEvent, app, enqueue_kernel_termination_status_update
    )
    event_dispatcher.subscribe(
        SessionTerminatingBroadcastEvent, app, enqueue_session_termination_status_update
    )
    event_dispatcher.subscribe(
        SessionTerminatedBroadcastEvent, app, enqueue_session_termination_status_update
    )
    event_dispatcher.subscribe(
        SessionCancelledBroadcastEvent, app, enqueue_session_creation_status_update
    )
    event_dispatcher.subscribe(SessionSuccessBroadcastEvent, app, enqueue_batch_task_result_update)
    event_dispatcher.subscribe(SessionFailureBroadcastEvent, app, enqueue_batch_task_result_update)
    yield


async def events_shutdown(app: web.Application) -> None:
    # shutdown handler is called before waiting for closing active connections.
    # We need to put sentinels here to ensure delivery of them to active SSE connections.
    app_ctx: PrivateContext = app["events.context"]
    join_tasks = []
    for sq in app_ctx.session_event_queues:
        sq.put_nowait(sentinel)
        join_tasks.append(sq.join())
    await asyncio.gather(*join_tasks)


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "events"
    app["events.context"] = PrivateContext()
    app["api_versions"] = (3, 4)
    app.on_shutdown.append(events_shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    app.cleanup_ctx.append(events_app_ctx)
    cors.add(add_route("GET", r"/background-task", push_background_task_events))
    cors.add(add_route("GET", r"/session", push_session_events))
    return app, []



import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, override

from aiohttp import web

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.config.unified import AuthConfig
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class UpdatePasswordNoAuthAction(AuthAction):
    request: web.Request
    domain_name: str
    email: str
    current_password: str
    new_password: str
    auth_config: Optional[AuthConfig]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_password_no_auth"

    @property
    def hook_params(self) -> dict[str, str]:
        return {
            "domain": self.domain_name,
            "username": self.email,
            "current_password": self.current_password,
            "new_password": self.new_password,
        }


@dataclass
class UpdatePasswordNoAuthActionResult(BaseActionResult):
    user_id: uuid.UUID
    password_changed_at: datetime

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_id)



from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import ImageAliasData
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class AliasImageAction(ImageAction):
    image_canonical: str
    architecture: str
    alias: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "alias"


@dataclass
class AliasImageActionResult(BaseActionResult):
    image_id: UUID
    image_alias: ImageAliasData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image_id)


class AliasImageActionValueError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-parameters"
    error_title = "Invalid parameters for image alias."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE_ALIAS,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class AliasImageActionDBError(BackendAIError, web.HTTPInternalServerError):
    """
    This can occur when an image alias with the same value already exists.
    """

    error_type = "https://api.backend.ai/probs/image-db-error"
    error_title = "Database error while managing image alias."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE_ALIAS,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )



"""
This module defines a series of Backend.AI-specific errors based on HTTP Error
classes from aiohttp.
Raising a BackendError is automatically mapped to a corresponding HTTP error
response with RFC7807-style JSON-encoded description in its response body.

In the client side, you should use "type" field in the body to distinguish
canonical error types beacuse "title" field may change due to localization and
future UX improvements.
"""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError as BackendError,
)
from ai.backend.common.exception import (
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.common.json import dump_json
from ai.backend.common.plugin.hook import HookResult

from ..exceptions import AgentError

if TYPE_CHECKING:
    from ai.backend.manager.api.vfolder import VFolderRow


class URLNotFound(BackendError, web.HTTPNotFound):  # TODO: Misused now.
    error_type = "https://api.backend.ai/probs/url-not-found"
    error_title = "Unknown URL path."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ObjectNotFound(BackendError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/object-not-found"
    object_name = "object"

    def __init__(
        self,
        extra_msg: Optional[str] = None,
        extra_data: Optional[Any] = None,
        *,
        object_name: Optional[str] = None,
        **kwargs,
    ) -> None:
        if object_name:
            self.object_name = object_name
        self.error_title = f"No such {self.object_name}."
        super().__init__(extra_msg, extra_data, **kwargs)

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class GenericBadRequest(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Bad request."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class KernelNotReady(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/kernel-not-ready"
    error_title = "Kernel not ready."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_READY,
        )


class RejectedByHook(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/rejected-by-hook"
    error_title = "Operation rejected by a hook plugin."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.PLUGIN,
            operation=ErrorOperation.HOOK,
            error_detail=ErrorDetail.FORBIDDEN,
        )

    @classmethod
    def from_hook_result(cls, result: HookResult) -> RejectedByHook:
        return cls(
            extra_msg=result.reason,
            extra_data={
                "plugins": result.src_plugin,
            },
        )


class InvalidCredentials(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-credentials"
    error_title = "Invalid credentials for authentication."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class GenericForbidden(BackendError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/generic-forbidden"
    error_title = "Forbidden operation."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class InsufficientPrivilege(BackendError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/insufficient-privilege"
    error_title = "Insufficient privilege."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class MethodNotAllowed(BackendError, web.HTTPMethodNotAllowed):
    error_type = "https://api.backend.ai/probs/method-not-allowed"
    error_title = "HTTP Method Not Allowed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class InternalServerError(BackendError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Internal server error."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ServerMisconfiguredError(BackendError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/server-misconfigured"
    error_title = "Service misconfigured."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ServiceUnavailable(BackendError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/service-unavailable"
    error_title = "Serivce unavailable."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class NotImplementedAPI(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/not-implemented"
    error_title = "This API is not implemented."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.NOT_IMPLEMENTED,
        )


class DeprecatedAPI(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/deprecated"
    error_title = "This API is deprecated."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.DEPRECATED,
        )


class InvalidAuthParameters(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-auth-params"
    error_title = "Missing or invalid authorization parameters."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class AuthorizationFailed(BackendError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/auth-failed"
    error_title = "Credential/signature mismatch."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.UNAUTHORIZED,
        )


class PasswordExpired(BackendError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/password-expired"
    error_title = "Password has expired."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.DATA_EXPIRED,
        )


class InvalidAPIParameters(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-api-params"
    error_title = "Missing or invalid API parameters."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class GraphQLError(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/graphql-error"
    error_title = "GraphQL-generated error."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ContainerRegistryWebhookAuthorizationFailed(BackendError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/webhook/auth-failed"
    error_title = "Container Registry Webhook authorization failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.CONTAINER_REGISTRY,
            operation=ErrorOperation.HOOK,
            error_detail=ErrorDetail.UNAUTHORIZED,
        )


class HarborWebhookContainerRegistryRowNotFound(InternalServerError):
    error_type = "https://api.backend.ai/probs/webhook/harbor/container-registry-not-found"
    error_title = "Container registry row not found."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.CONTAINER_REGISTRY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class InstanceNotFound(ObjectNotFound):
    object_name = "agent instance"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.INSTANCE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ImageNotFound(ObjectNotFound):
    object_name = "environment image"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class DomainNotFound(ObjectNotFound):
    object_name = "domain"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DOMAIN,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class GroupNotFound(ObjectNotFound):
    object_name = "user group (or project)"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.GROUP,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class UserNotFound(ObjectNotFound):
    object_name = "user"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ScalingGroupNotFound(ObjectNotFound):
    object_name = "scaling group"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SCALING_GROUP,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class SessionNotFound(ObjectNotFound):
    object_name = "session"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class MainKernelNotFound(ObjectNotFound):
    object_name = "main kernel"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class KernelNotFound(ObjectNotFound):
    object_name = "kernel"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class EndpointNotFound(ObjectNotFound):
    object_name = "endpoint"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class RoutingNotFound(ObjectNotFound):
    object_name = "routing"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ROUTE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class EndpointTokenNotFound(ObjectNotFound):
    object_name = "endpoint_token"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ContainerRegistryNotFound(ObjectNotFound):
    object_name = "container_registry"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.CONTAINER_REGISTRY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ContainerRegistryGroupsAssociationNotFound(ObjectNotFound):
    object_name = "association of container_registry and group"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.CONTAINER_REGISTRY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class TooManySessionsMatched(BackendError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/too-many-sessions-matched"
    error_title = "Too many sessions matched."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.CONFLICT,
        )

    def __init__(
        self,
        extra_msg: Optional[str] = None,
        extra_data: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        if extra_data is not None and (matches := extra_data.get("matches", None)) is not None:
            serializable_matches = [
                {
                    "id": str(item["session_id"]),
                    "name": item["session_name"],
                    "status": item["status"].name,
                    "created_at": item["created_at"].isoformat(),
                }
                for item in matches
            ]
            extra_data["matches"] = serializable_matches
        super().__init__(extra_msg, extra_data, **kwargs)


class TooManyKernelsFound(BackendError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/too-many-kernels"
    error_title = "There are two or more matching kernels."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.CONFLICT,
        )


class TaskTemplateNotFound(ObjectNotFound):
    object_name = "task template"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.TEMPLATE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class AppNotFound(ObjectNotFound):
    object_name = "app service"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class SessionAlreadyExists(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/session-already-exists"
    error_title = "The session already exists but you requested not to reuse existing one."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class TooManyVFoldersFound(BackendError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/too-many-vfolders"
    error_title = "Multiple vfolders found for the operation for a single vfolder."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.CONFLICT,
        )

    def __init__(self, matched_rows: Sequence[VFolderRow]) -> None:
        serialized_matches = [
            {
                "id": row["id"],
                "host": row["host"],
                "user": row["user_email"],
                "user_id": row["user"],
                "group": row["group_name"],
                "group_id": row["group"],
            }
            for row in matched_rows
        ]
        super().__init__(extra_data={"matches": serialized_matches})


class VFolderNotFound(ObjectNotFound):
    object_name = "virtual folder"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class VFolderAlreadyExists(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-already-exists"
    error_title = "The virtual folder already exists with the same name."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class VFolderGone(BackendError, web.HTTPGone):
    error_type = "https://api.backend.ai/probs/vfolder-gone"
    error_title = "The virtual folder is gone."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.GONE,
        )


class ModelServiceDependencyNotCleared(BackendError, web.HTTPBadRequest):
    error_title = "Cannot delete model VFolders bound to alive model services."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.CONFLICT,
        )


class VFolderOperationFailed(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-operation-failed"
    error_title = "Virtual folder operation has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class VFolderFilterStatusFailed(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-filter-status-failed"
    error_title = "Virtual folder status filtering has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class VFolderFilterStatusNotAvailable(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-filter-status-not-available"
    error_title = "There is no available virtual folder to filter its status."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class VFolderPermissionError(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-permission-error"
    error_title = "The virtual folder does not permit the specified permission."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class DotfileCreationFailed(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Dotfile creation has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DOTFILE,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class DotfileAlreadyExists(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Dotfile already exists."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DOTFILE,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class DotfileNotFound(ObjectNotFound):
    object_name = "dotfile"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DOTFILE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class DotfileVFolderPathConflict(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/dotfile-vfolder-path-conflict"
    error_title = "The dotfile path conflicts with a virtual folder path."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DOTFILE,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.CONFLICT,
        )


class QuotaExceeded(BackendError, web.HTTPPreconditionFailed):
    error_type = "https://api.backend.ai/probs/quota-exceeded"
    error_title = "You have reached your resource limit."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class RateLimitExceeded(BackendError, web.HTTPTooManyRequests):
    error_type = "https://api.backend.ai/probs/rate-limit-exceeded"
    error_title = "You have reached your API query rate limit."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class InstanceNotAvailable(BackendError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/instance-not-available"
    error_title = "There is no available instance."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.INSTANCE,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class ServerFrozen(BackendError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/server-frozen"
    error_title = "The server is frozen due to maintenance. Please try again later."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class StorageProxyError(BackendError, web.HTTPError):
    error_type = "https://api.backend.ai/probs/storage-proxy-error"
    error_title = "The storage proxy returned an error."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )

    def __init__(self, status: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Currently there is no good public way to override the status code
        # after initialization of aiohttp.web.StreamResponse objects. :(
        self.status_code = status  # HTTPException uses self.status_code
        self._status = status  # StreamResponse uses self._status
        self.args = (status, self.args[1], self.args[2])

    @property
    def status(self) -> int:
        # override the status property again to refer the subclass' attribute.
        return self.status_code


class BackendAgentError(BackendError):
    """
    An RFC-7807 error class that wraps agent-side errors.
    """

    _short_type_map = {
        "TIMEOUT": "https://api.backend.ai/probs/agent-timeout",
        "INVALID_INPUT": "https://api.backend.ai/probs/agent-invalid-input",
        "FAILURE": "https://api.backend.ai/probs/agent-failure",
    }

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.AGENT,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )

    def __init__(
        self,
        agent_error_type: str,
        exc_info: Union[str, AgentError, Exception, Mapping[str, Optional[str]], None] = None,
    ):
        super().__init__()
        agent_details: Mapping[str, Optional[str]]
        if not agent_error_type.startswith("https://"):
            agent_error_type = self._short_type_map[agent_error_type.upper()]
        self.args = (
            self.status_code,
            self.reason,
            self.error_type,
            agent_error_type,
        )
        if isinstance(exc_info, str):
            agent_details = {
                "type": agent_error_type,
                "title": exc_info,
            }
        elif isinstance(exc_info, AgentError):
            e = cast(AgentError, exc_info)
            agent_details = {
                "type": agent_error_type,
                "title": "Agent-side exception occurred.",
                "exception": e.exc_repr,
            }
        elif isinstance(exc_info, Exception):
            agent_details = {
                "type": agent_error_type,
                "title": "Unexpected exception ocurred.",
                "exception": repr(exc_info),
            }
        elif isinstance(exc_info, Mapping):
            agent_details = exc_info
        else:
            agent_details = {
                "type": agent_error_type,
                "title": None if exc_info is None else str(exc_info),
            }
        self.agent_details = agent_details
        self.agent_error_type = agent_error_type
        self.agent_error_title = agent_details["title"]
        self.agent_exception = agent_details.get("exception", "")
        self.body = dump_json({
            "type": self.error_type,
            "title": self.error_title,
            "agent-details": agent_details,
        })

    def __str__(self):
        if self.agent_exception:
            return f"{self.agent_error_title} ({self.agent_exception})"
        return f"{self.agent_error_title}"

    def __repr__(self):
        if self.agent_exception:
            return f"<{type(self).__name__}: {self.agent_error_title} ({self.agent_exception})>"
        return f"<{type(self).__name__}: {self.agent_error_title}>"

    def __reduce__(self):
        return (type(self), (self.agent_error_type, self.agent_details))


class KernelCreationFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-creation-failed"
    error_title = "Kernel creation has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KernelDestructionFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-destruction-failed"
    error_title = "Kernel destruction has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KernelRestartFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-restart-failed"
    error_title = "Kernel restart has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.START,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KernelExecutionFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-execution-failed"
    error_title = "Executing user code in the kernel has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UnknownImageReferenceError(ObjectNotFound):
    object_name = "image reference"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )



from ai.backend.manager.config.provider import ManagerConfigProvider

from .container_metric import ContainerUtilizationMetricService


class UtilizationMetricService:
    container: ContainerUtilizationMetricService

    def __init__(self, config_provider: ManagerConfigProvider) -> None:
        self.container = ContainerUtilizationMetricService(config_provider)



import uuid
from pathlib import PurePosixPath
from typing import (
    Any,
    Optional,
    cast,
)

import aiohttp
import sqlalchemy as sa
from aiohttp import hdrs, web
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.defs import VFOLDER_GROUP_PERMISSION_MODE
from ai.backend.common.types import (
    QuotaScopeID,
    QuotaScopeType,
    VFolderHostPermission,
    VFolderID,
    VFolderUsageMode,
)
from ai.backend.manager.config.constant import DEFAULT_CHUNK_SIZE
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.errors.exceptions import ObjectNotFound, StorageProxyError
from ai.backend.manager.models.endpoint import EndpointLifecycle, EndpointRow
from ai.backend.manager.models.group import GroupRow, ProjectType
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole, UserRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_txn_retry
from ai.backend.manager.models.vfolder import (
    HARD_DELETED_VFOLDER_STATUSES,
    VFolderCloneInfo,
    VFolderDeletionInfo,
    VFolderOperationStatus,
    VFolderOwnershipType,
    VFolderPermission,
    VFolderPermissionRow,
    VFolderRow,
    VFolderStatusSet,
    delete_vfolder_relation_rows,
    ensure_host_permission_allowed,
    filter_host_allowed_permission,
    initiate_vfolder_clone,
    initiate_vfolder_deletion,
    is_unmanaged,
    query_accessible_vfolders,
    update_vfolder_status,
    verify_vfolder_name,
    vfolder_status_map,
    vfolders,
)

# from ai.backend.manager.types import SENTINEL
from ..actions.base import (
    CloneVFolderAction,
    CloneVFolderActionResult,
    CreateVFolderAction,
    CreateVFolderActionResult,
    DeleteForeverVFolderAction,
    DeleteForeverVFolderActionResult,
    ForceDeleteVFolderAction,
    ForceDeleteVFolderActionResult,
    GetTaskLogsAction,
    GetTaskLogsActionResult,
    GetVFolderAction,
    GetVFolderActionResult,
    ListVFolderAction,
    ListVFolderActionResult,
    MoveToTrashVFolderAction,
    MoveToTrashVFolderActionResult,
    RestoreVFolderFromTrashAction,
    RestoreVFolderFromTrashActionResult,
    UpdateVFolderAttributeAction,
    UpdateVFolderAttributeActionResult,
)
from ..exceptions import (
    Forbidden,
    ModelServiceDependencyNotCleared,
    ProjectNotFound,
    TooManyVFoldersFound,
    VFolderAlreadyExists,
    VFolderCreationFailure,
    VFolderFilterStatusFailed,
    VFolderFilterStatusNotAvailable,
    VFolderInvalidParameter,
    VFolderNotFound,
)
from ..types import VFolderBaseInfo, VFolderOwnershipInfo, VFolderUsageInfo


async def _check_vfolder_status(
    vfolder_status: VFolderOperationStatus,
    status: VFolderStatusSet,
) -> None:
    """
    Checks if the target vfolder status matches one of the status sets aliased by `status` VFolderStatusSet,
    """

    available_vf_statuses = vfolder_status_map.get(status)
    if not available_vf_statuses:
        raise VFolderFilterStatusNotAvailable
    if vfolder_status not in available_vf_statuses:
        raise VFolderFilterStatusFailed


class VFolderService:
    _db: ExtendedAsyncSAEngine
    _config_provider: ManagerConfigProvider
    _storage_manager: StorageSessionManager
    _background_task_manager: BackgroundTaskManager

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        config_provider: ManagerConfigProvider,
        storage_manager: StorageSessionManager,
        background_task_manager: BackgroundTaskManager,
    ) -> None:
        self._db = db
        self._config_provider = config_provider
        self._storage_manager = storage_manager
        self._background_task_manager = background_task_manager

    async def create(self, action: CreateVFolderAction) -> CreateVFolderActionResult:
        user_role = action.user_role
        user_uuid = action.user_uuid
        keypair_resource_policy = action.keypair_resource_policy
        domain_name = action.domain_name
        group_id_or_name = action.group_id_or_name
        folder_host = action.folder_host
        unmanaged_path = action.unmanaged_path
        # Resolve host for the new virtual folder.
        if not folder_host:
            folder_host = self._config_provider.config.volumes.default_host
            if not folder_host:
                raise VFolderInvalidParameter(
                    "You must specify the vfolder host because the default host is not configured."
                )
        # Check if user is trying to created unmanaged vFolder
        if unmanaged_path:
            # Approve only if user is Admin or Superadmin
            if user_role not in (UserRole.ADMIN, UserRole.SUPERADMIN):
                raise Forbidden("Insufficient permission")
                # Assign ghost host to unmanaged vfolder

        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )

        if action.name.startswith(".") and action.name != ".local":
            if action.group_id_or_name is not None:
                raise VFolderInvalidParameter("dot-prefixed vfolders cannot be a group folder.")

        group_uuid: Optional[uuid.UUID] = None
        group_type: Optional[ProjectType] = None
        max_vfolder_count: int
        max_quota_scope_size: int
        container_uid: Optional[int] = None

        async with self._db.begin_session() as sess:
            match group_id_or_name:
                case str():
                    # Convert the group name to group uuid.
                    # log.debug("group_id_or_name(str):{}", group_id_or_name)
                    query = (
                        sa.select(GroupRow)
                        .where(
                            (GroupRow.domain_name == domain_name)
                            & (GroupRow.name == group_id_or_name)
                        )
                        .options(selectinload(GroupRow.resource_policy_row))
                    )
                    result = await sess.execute(query)
                    group_row = cast(Optional[GroupRow], result.scalar())
                    if group_row is None:
                        raise ProjectNotFound(group_id_or_name)
                    _gid, max_vfolder_count, max_quota_scope_size = (
                        cast(Optional[uuid.UUID], group_row.id),
                        cast(int, group_row.resource_policy_row.max_vfolder_count),
                        cast(int, group_row.resource_policy_row.max_quota_scope_size),
                    )
                    if _gid is None:
                        raise ProjectNotFound(group_id_or_name)
                    group_uuid = _gid
                    group_type = cast(ProjectType, group_row.type)
                case uuid.UUID():
                    # Check if the group belongs to the current domain.
                    # log.debug("group_id_or_name(uuid):{}", group_id_or_name)
                    query = (
                        sa.select(GroupRow)
                        .where(
                            (GroupRow.domain_name == domain_name)
                            & (GroupRow.id == group_id_or_name)
                        )
                        .options(selectinload(GroupRow.resource_policy_row))
                    )
                    result = await sess.execute(query)
                    group_row = cast(Optional[GroupRow], result.scalar())
                    if group_row is None:
                        raise ProjectNotFound(group_id_or_name)
                    _gid, max_vfolder_count, max_quota_scope_size = (
                        group_row.id,
                        cast(int, group_row.resource_policy_row.max_vfolder_count),
                        cast(int, group_row.resource_policy_row.max_quota_scope_size),
                    )
                    if _gid is None:
                        raise ProjectNotFound(group_id_or_name)
                    group_uuid = group_id_or_name
                    group_type = cast(ProjectType, group_row.type)
                case None:
                    query = (
                        sa.select(UserRow)
                        .where(UserRow.uuid == user_uuid)
                        .options(selectinload(UserRow.resource_policy_row))
                    )
                    result = await sess.execute(query)
                    user_row = result.scalar()
                    max_vfolder_count, max_quota_scope_size = (
                        cast(int, user_row.resource_policy_row.max_vfolder_count),
                        cast(int, user_row.resource_policy_row.max_quota_scope_size),
                    )
                    container_uid = cast(Optional[int], user_row.container_uid)
                case _:
                    raise ProjectNotFound(group_id_or_name)

            vfolder_permission_mode = (
                VFOLDER_GROUP_PERMISSION_MODE if container_uid is not None else None
            )

            # Check if group exists when it's given a non-empty value.
            if group_id_or_name and group_uuid is None:
                raise ProjectNotFound(group_id_or_name)

            # Determine the ownership type and the quota scope ID.
            if group_uuid is not None:
                ownership_type = "group"
                quota_scope_id = QuotaScopeID(QuotaScopeType.PROJECT, group_uuid)
                if (
                    user_role not in (UserRole.SUPERADMIN, UserRole.ADMIN)
                    and group_type != ProjectType.MODEL_STORE
                ):
                    raise Forbidden("no permission")
            else:
                ownership_type = "user"
                quota_scope_id = QuotaScopeID(QuotaScopeType.USER, user_uuid)
            if ownership_type not in allowed_vfolder_types:
                raise VFolderInvalidParameter(
                    f"{ownership_type}-owned vfolder is not allowed in this cluster"
                )

        if group_type == ProjectType.MODEL_STORE:
            if action.mount_permission != VFolderPermission.READ_WRITE:
                raise VFolderInvalidParameter(
                    "Setting custom permission is not supported for model store vfolder"
                )
            if action.usage_mode != VFolderUsageMode.MODEL:
                raise VFolderInvalidParameter(
                    "Only Model VFolder can be created under the model store project"
                )

        async with self._db.begin() as conn:
            await ensure_host_permission_allowed(
                conn,
                folder_host,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=user_uuid,
                resource_policy=keypair_resource_policy,
                domain_name=domain_name,
                group_id=group_uuid,
                permission=VFolderHostPermission.CREATE,
            )

            # Check resource policy's max_vfolder_count
            if max_vfolder_count > 0:
                if ownership_type == "user":
                    query = (
                        sa.select(sa.func.count())
                        .select_from(VFolderRow)
                        .where(
                            (VFolderRow.user == user_uuid)
                            & (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES))
                        )
                    )
                else:
                    assert group_uuid is not None
                    query = (
                        sa.select(sa.func.count())
                        .select_from(VFolderRow)
                        .where(
                            (VFolderRow.group == group_uuid)
                            & (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES))
                        )
                    )
                result = cast(int, await conn.scalar(query))
                if result >= max_vfolder_count:
                    raise VFolderInvalidParameter("You cannot create more vfolders.")

            # DEPRECATED: Limit vfolder size quota if it is larger than max_vfolder_size of the resource policy.
            # max_vfolder_size = resource_policy.get("max_vfolder_size", 0)
            # if max_vfolder_size > 0 and (
            #     params["quota"] is None or params["quota"] <= 0 or params["quota"] > max_vfolder_size
            # ):
            #     params["quota"] = max_vfolder_size

            # Prevent creation of vfolder with duplicated name on all hosts.
            extra_vf_conds = [
                (VFolderRow.name == action.name),
                (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES)),
            ]
            entries = await query_accessible_vfolders(
                conn,
                user_uuid,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=(sa.and_(*extra_vf_conds)),
            )
            if len(entries) > 0:
                raise VFolderAlreadyExists(
                    f"VFolder with the given name already exists. ({action.name})"
                )
            try:
                folder_id = uuid.uuid4()
                vfid = VFolderID(quota_scope_id, folder_id)
                if not unmanaged_path:
                    options = {}
                    if max_quota_scope_size and max_quota_scope_size > 0:
                        options["initial_max_size_for_quota_scope"] = max_quota_scope_size
                    body_data: dict[str, Any] = {
                        "volume": self._storage_manager.get_proxy_and_volume(
                            folder_host, is_unmanaged(unmanaged_path)
                        )[1],
                        "vfid": str(vfid),
                        "options": options,
                    }
                    if vfolder_permission_mode is not None:
                        body_data["mode"] = vfolder_permission_mode
                    async with self._storage_manager.request(
                        folder_host,
                        "POST",
                        "folder/create",
                        json=body_data,
                    ):
                        pass
            except aiohttp.ClientResponseError as e:
                raise VFolderCreationFailure from e

            # By default model store VFolder should be considered as read only for every users but without the creator
            if group_type == ProjectType.MODEL_STORE:
                action.mount_permission = VFolderPermission.READ_ONLY

            # TODO: include quota scope ID in the database
            # TODO: include quota scope ID in the API response
            insert_values = {
                "id": vfid.folder_id.hex,
                "name": action.name,
                "domain_name": domain_name,
                "quota_scope_id": str(quota_scope_id),
                "usage_mode": action.usage_mode,
                "permission": action.mount_permission,
                "last_used": None,
                "host": folder_host,
                "creator": action.creator_email,
                "ownership_type": VFolderOwnershipType(ownership_type),
                "user": user_uuid if ownership_type == "user" else None,
                "group": group_uuid if ownership_type == "group" else None,
                "unmanaged_path": unmanaged_path,
                "cloneable": action.cloneable,
                "status": VFolderOperationStatus.READY,
            }
            try:
                query = sa.insert(VFolderRow, insert_values)
                result = await conn.execute(query)

                # Here we grant creator the permission to alter VFolder contents
                if group_type == ProjectType.MODEL_STORE:
                    query = sa.insert(VFolderPermissionRow).values({
                        "user": user_uuid,
                        "vfolder": vfid.folder_id.hex,
                        "permission": VFolderPermission.OWNER_PERM,
                    })
                    await conn.execute(query)
            except sa.exc.DataError:
                raise VFolderInvalidParameter
            assert result.rowcount == 1

        return CreateVFolderActionResult(
            id=vfid.folder_id,
            name=action.name,
            quota_scope_id=quota_scope_id,
            host=folder_host,
            unmanaged_path=unmanaged_path,
            usage_mode=action.usage_mode,
            mount_permission=action.mount_permission,
            user_uuid=user_uuid if ownership_type == "user" else None,
            group_uuid=group_uuid if ownership_type == "group" else None,
            creator_email=action.creator_email,
            ownership_type=VFolderOwnershipType(ownership_type),
            cloneable=action.cloneable,
            status=VFolderOperationStatus.READY,
        )

    async def update_attribute(
        self, action: UpdateVFolderAttributeAction
    ) -> UpdateVFolderAttributeActionResult:
        modifier = action.modifier
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )

        async def _update(db_session: AsyncSession) -> None:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            requester_user_row = cast(UserRow, requester_user_row)
            vfolder_dicts = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
            )
            if not vfolder_dicts:
                raise VFolderNotFound
            query_vfolder = sa.select(VFolderRow).where(VFolderRow.id == action.vfolder_uuid)
            vfolder_row = await db_session.scalar(query_vfolder)
            vfolder_row = cast(VFolderRow, vfolder_row)
            try:
                new_name = modifier.name.value()
            except ValueError:
                pass
            else:
                for row in vfolder_dicts:
                    if row["name"] == new_name:
                        raise VFolderInvalidParameter(
                            "One of your accessible vfolders already has the name you requested."
                        )
            to_update = modifier.fields_to_update()
            for key, value in to_update.items():
                setattr(vfolder_row, key, value)

        async with self._db.connect() as db_conn:
            await execute_with_txn_retry(_update, self._db.begin_session, db_conn)
        return UpdateVFolderAttributeActionResult(vfolder_uuid=action.vfolder_uuid)

    async def get(self, action: GetVFolderAction) -> GetVFolderActionResult:
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            requester_user_row = cast(UserRow, requester_user_row)
            vfolder_dicts = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )
            vfolder_row = vfolder_dicts[0]
        if vfolder_row["permission"] is None:
            is_owner = True
            permission = VFolderPermission.OWNER_PERM
        else:
            is_owner = vfolder_row["is_owner"]
            permission = vfolder_row["permission"]
        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_row["host"], is_unmanaged(vfolder_row["unmanaged_path"])
        )
        async with self._storage_manager.request(
            proxy_name,
            "GET",
            "folder/usage",
            json={
                "volume": volume_name,
                "vfid": str(VFolderID(vfolder_row["quota_scope_id"], vfolder_row["id"])),
            },
        ) as (_, storage_resp):
            usage = await storage_resp.json()
            usage_info = VFolderUsageInfo(
                used_bytes=usage["used_bytes"],
                num_files=usage["file_count"],
            )
        return GetVFolderActionResult(
            user_uuid=action.user_uuid,
            base_info=VFolderBaseInfo(
                id=vfolder_row["id"],
                quota_scope_id=vfolder_row["quota_scope_id"],
                name=vfolder_row["name"],
                host=vfolder_row["host"],
                status=vfolder_row["status"],
                unmanaged_path=vfolder_row["unmanaged_path"],
                mount_permission=permission,
                usage_mode=vfolder_row["usage_mode"],
                created_at=vfolder_row["created_at"],
                cloneable=vfolder_row["cloneable"],
            ),
            ownership_info=VFolderOwnershipInfo(
                creator_email=vfolder_row["creator"],
                ownership_type=vfolder_row["ownership_type"],
                is_owner=is_owner,
                user_uuid=vfolder_row["user"],
                group_uuid=vfolder_row["group"],
            ),
            usage_info=usage_info,
        )

    async def list(self, action: ListVFolderAction) -> ListVFolderActionResult:
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            requester_user_row = cast(UserRow, requester_user_row)
            vfolder_dicts = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=False,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
            )
        vfolders = [
            (
                VFolderBaseInfo(
                    id=entry["id"],
                    quota_scope_id=entry["quota_scope_id"],
                    name=entry["name"],
                    host=entry["host"],
                    status=entry["status"],
                    unmanaged_path=entry["unmanaged_path"],
                    mount_permission=entry["permission"],
                    usage_mode=entry["usage_mode"],
                    created_at=entry["created_at"],
                    cloneable=entry["cloneable"],
                ),
                VFolderOwnershipInfo(
                    creator_email=entry["creator"],
                    ownership_type=entry["ownership_type"],
                    is_owner=entry["is_owner"],
                    user_uuid=entry["user"],
                    group_uuid=entry["group"],
                ),
            )
            for entry in vfolder_dicts
        ]

        return ListVFolderActionResult(
            user_uuid=action.user_uuid,
            vfolders=vfolders,
        )

    async def move_to_trash(
        self, action: MoveToTrashVFolderAction
    ) -> MoveToTrashVFolderActionResult:
        # Only the effective folder owner can delete the folder.
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        async with self._db.connect() as db_conn:
            async with self._db.begin_session(db_conn) as db_session:
                requester_user_row = await db_session.scalar(
                    sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
                )
                requester_user_row = cast(UserRow, requester_user_row)
                vfolder_dicts = await query_accessible_vfolders(
                    db_session.bind,
                    action.user_uuid,
                    allow_privileged_access=True,
                    user_role=requester_user_row.role,
                    allowed_vfolder_types=allowed_vfolder_types,
                    domain_name=requester_user_row.domain_name,
                    extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
                )
                vfolder_row = vfolder_dicts[0]
                if not vfolder_row["is_owner"]:
                    raise VFolderInvalidParameter(
                        "Cannot delete the vfolder that is not owned by myself."
                    )
                await _check_vfolder_status(vfolder_row["status"], VFolderStatusSet.DELETABLE)
                # perform extra check to make sure records of alive model service not removed by foreign key rule
                if vfolder_row["usage_mode"] == VFolderUsageMode.MODEL:
                    live_endpoints = await EndpointRow.list_by_model(db_session, vfolder_row["id"])
                    if (
                        len([
                            e
                            for e in live_endpoints
                            if e.lifecycle_stage == EndpointLifecycle.CREATED
                        ])
                        > 0
                    ):
                        raise ModelServiceDependencyNotCleared
                folder_host = vfolder_row["host"]
                await ensure_host_permission_allowed(
                    db_session.bind,
                    folder_host,
                    allowed_vfolder_types=allowed_vfolder_types,
                    user_uuid=action.user_uuid,
                    resource_policy=action.keypair_resource_policy,
                    domain_name=requester_user_row.domain_name,
                    permission=VFolderHostPermission.DELETE,
                )

                vfolder_row_ids = (vfolder_row["id"],)
            await delete_vfolder_relation_rows(db_conn, self._db.begin_session, vfolder_row_ids)
        await update_vfolder_status(
            self._db,
            vfolder_row_ids,
            VFolderOperationStatus.DELETE_PENDING,
        )
        return MoveToTrashVFolderActionResult(vfolder_uuid=action.vfolder_uuid)

    async def restore(
        self, action: RestoreVFolderFromTrashAction
    ) -> RestoreVFolderFromTrashActionResult:
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )

        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            restore_targets = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )

            if len(restore_targets) > 1:
                raise TooManyVFoldersFound(restore_targets)
            elif len(restore_targets) == 0:
                raise VFolderInvalidParameter("No such vfolder.")

            row = restore_targets[0]
            await _check_vfolder_status(row["status"], VFolderStatusSet.RECOVERABLE)

        # Folder owner OR user who have DELETE permission can restore folder.
        if not row["is_owner"] and row["permission"] != VFolderPermission.RW_DELETE:
            raise VFolderInvalidParameter("Cannot restore the vfolder that is not owned by myself.")

        # fs-level mv may fail or take longer time
        # but let's complete the db transaction to reflect that it's deleted.
        await update_vfolder_status(self._db, (row["id"],), VFolderOperationStatus.READY)
        return RestoreVFolderFromTrashActionResult(vfolder_uuid=action.vfolder_uuid)

    async def delete_forever(
        self, action: DeleteForeverVFolderAction
    ) -> DeleteForeverVFolderActionResult:
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )

        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            entries = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )

            if len(entries) > 1:
                raise TooManyVFoldersFound(entries)
            elif len(entries) == 0:
                raise VFolderInvalidParameter("No such vfolder.")
            row = entries[0]
            await _check_vfolder_status(row["status"], VFolderStatusSet.PURGABLE)

        # fs-level deletion may fail or take longer time
        await initiate_vfolder_deletion(
            self._db,
            [VFolderDeletionInfo(VFolderID.from_row(row), row["host"], row["unmanaged_path"])],
            self._storage_manager,
        )
        return DeleteForeverVFolderActionResult(vfolder_uuid=action.vfolder_uuid)

    async def force_delete(
        self, action: ForceDeleteVFolderAction
    ) -> ForceDeleteVFolderActionResult:
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )

        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            entries = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )
            row = entries[0]
            try:
                await _check_vfolder_status(row["status"], VFolderStatusSet.PURGABLE)
            except VFolderFilterStatusFailed:
                await _check_vfolder_status(row["status"], VFolderStatusSet.DELETABLE)
        await initiate_vfolder_deletion(
            self._db,
            [VFolderDeletionInfo(VFolderID.from_row(row), row["host"], row["unmanaged_path"])],
            storage_manager=self._storage_manager,
            force=True,
        )
        return ForceDeleteVFolderActionResult(vfolder_uuid=action.vfolder_uuid)

    async def clone(self, action: CloneVFolderAction) -> CloneVFolderActionResult:
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        if "user" not in allowed_vfolder_types:
            raise VFolderInvalidParameter("user vfolder cannot be created in this host")
        requester_user_row = await UserRow.get_by_id_with_policies(
            action.requester_user_uuid, db=self._db
        )
        if requester_user_row is None:
            raise VFolderInvalidParameter("No such user.")
        async with self._db.begin_readonly_session() as db_session:
            entries = await query_accessible_vfolders(
                db_session.bind,
                action.requester_user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
            )
            for entry in entries:
                if entry["id"] == action.source_vfolder_uuid:
                    row = entry
                    break
            else:
                raise VFolderInvalidParameter("No such vfolder.")
        project_row: Optional[GroupRow] = None
        if (project_id := row["group"]) is not None:
            project_row = await GroupRow.get_by_id_with_policies(project_id, db=self._db)
        async with self._db.begin_session() as db_session:
            domain_name = requester_user_row.domain_name
            source_folder_host = row["host"]
            source_folder_id = VFolderID(row["quota_scope_id"], row["id"])
            target_folder_host = action.target_host or source_folder_host
            target_quota_scope_id = "..."  # TODO: implement
            source_unmanaged_path = row["unmanaged_path"]
            source_proxy_name, source_volume_name = self._storage_manager.get_proxy_and_volume(
                source_folder_host, is_unmanaged(source_unmanaged_path)
            )
            target_proxy_name, target_volume_name = self._storage_manager.get_proxy_and_volume(
                target_folder_host,
            )

            # check if the source vfolder is allowed to be cloned
            if not row["cloneable"]:
                raise Forbidden("The source vfolder is not permitted to be cloned.")

            if action.target_name.startswith("."):
                for entry in entries:
                    if entry["name"] == action.target_name:
                        raise VFolderAlreadyExists
                    if entry["name"].startswith(".") and entry["path"] == action.target_name:
                        raise VFolderInvalidParameter("VFolder name conflicts with your dotfile.")

            if not target_folder_host:
                target_folder_host = self._config_provider.config.volumes.default_host
                if not target_folder_host:
                    raise VFolderInvalidParameter(
                        "You must specify the vfolder host because the default host is not configured."
                    )

            if not verify_vfolder_name(action.target_name):
                raise VFolderInvalidParameter(
                    f"{action.target_name} is reserved for internal operations."
                )

            if source_proxy_name != target_proxy_name:
                raise VFolderInvalidParameter(
                    "proxy name of source and target vfolders must be equal."
                )

            if project_row is not None:
                vfolder_hosts = project_row.allowed_vfolder_hosts
                max_vfolder_count = project_row.resource_policy_row.max_vfolder_count
            else:
                vfolder_hosts = (
                    requester_user_row.main_keypair.resource_policy_row.allowed_vfolder_hosts
                )
                max_vfolder_count = requester_user_row.resource_policy_row.max_vfolder_count

            allowed_hosts = await filter_host_allowed_permission(
                db_session.bind,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=requester_user_row.uuid,
                resource_policy={"allowed_vfolder_hosts": vfolder_hosts},
                domain_name=domain_name,
            )
            if (
                target_folder_host not in allowed_hosts
                or VFolderHostPermission.CREATE not in allowed_hosts[target_folder_host]
            ):
                raise VFolderInvalidParameter(
                    f"`{VFolderHostPermission.CREATE}` Not allowed in vfolder"
                    f" host(`{target_folder_host}`)"
                )
            # TODO: handle legacy host lists assuming that volume names don't overlap?
            if target_folder_host not in allowed_hosts:
                raise VFolderInvalidParameter("You are not allowed to use this vfolder host.")

            # Check resource policy's max_vfolder_count
            if max_vfolder_count > 0:
                query = sa.select(sa.func.count()).where(
                    sa.and_(
                        VFolderRow.user == requester_user_row.uuid,
                        VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES),
                    )
                )
                result = await db_session.scalar(query)
                if result >= max_vfolder_count:
                    raise VFolderInvalidParameter("You cannot create more vfolders.")

        task_id, target_folder_id = await initiate_vfolder_clone(
            self._db,
            VFolderCloneInfo(
                source_folder_id,
                source_folder_host,
                source_unmanaged_path,
                domain_name,
                target_quota_scope_id,
                action.target_name,
                target_folder_host,
                action.usage_mode,
                action.mount_permission,
                requester_user_row.email,
                requester_user_row.uuid,
                action.cloneable,
            ),
            self._storage_manager,
            self._background_task_manager,
        )

        # Return the information about the destination vfolder.
        return CloneVFolderActionResult(
            vfolder_uuid=source_folder_id.folder_id,
            target_vfolder_id=target_folder_id,
            target_vfolder_name=action.target_name,
            target_vfolder_host=target_folder_host,
            usage_mode=action.usage_mode,
            mount_permission=action.mount_permission,
            creator_email=requester_user_row.email,
            ownership_type=VFolderOwnershipType.USER,
            owner_user_uuid=requester_user_row.uuid,
            owner_group_uuid=None,
            cloneable=action.cloneable,
            bgtask_id=task_id,
        )

    async def get_task_logs(self, action: GetTaskLogsAction) -> GetTaskLogsActionResult:
        user_uuid = action.user_id
        user_role = action.user_role
        domain_name = action.domain_name
        kernel_id_str = action.kernel_id.hex
        request = action.request

        async with self._db.begin_readonly() as conn:
            matched_vfolders = await query_accessible_vfolders(
                conn,
                user_uuid,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=["user"],
                extra_vf_conds=(vfolders.c.name == ".logs"),
            )
            if not matched_vfolders:
                raise ObjectNotFound(
                    extra_data={"vfolder_name": ".logs"},
                    object_name="vfolder",
                )
            log_vfolder = matched_vfolders[0]

        _proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            log_vfolder["host"], is_unmanaged(log_vfolder["unmanaged_path"])
        )
        response = web.StreamResponse(status=200)
        response.headers[hdrs.CONTENT_TYPE] = "text/plain"
        prepared = False

        try:
            async with self._storage_manager.request(
                log_vfolder["host"],
                "POST",
                "folder/file/fetch",
                json={
                    "volume": volume_name,
                    "vfid": str(VFolderID.from_row(log_vfolder)),
                    "relpath": str(
                        PurePosixPath("task")
                        / kernel_id_str[:2]
                        / kernel_id_str[2:4]
                        / f"{kernel_id_str[4:]}.log",
                    ),
                },
                raise_for_status=True,
            ) as (_, storage_resp):
                while True:
                    chunk = await storage_resp.content.read(DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    if not prepared:
                        await response.prepare(request)
                        prepared = True
                    await response.write(chunk)
        except aiohttp.ClientResponseError as e:
            raise StorageProxyError(status=e.status, extra_msg=e.message)
        finally:
            if prepared:
                await response.write_eof()
        return GetTaskLogsActionResult(response=response, vfolder_data=log_vfolder)



import logging
from io import BytesIO

import sqlalchemy as sa

from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.events.event_types.kernel.anycast import (
    DoSyncKernelLogsEvent,
    KernelCancelledAnycastEvent,
    KernelCreatingAnycastEvent,
    KernelHeartbeatEvent,
    KernelPreparingAnycastEvent,
    KernelPullingAnycastEvent,
    KernelStartedAnycastEvent,
    KernelTerminatedAnycastEvent,
    KernelTerminatingAnycastEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.registry import AgentRegistry

from ...models.kernel import kernels
from ...models.utils import (
    ExtendedAsyncSAEngine,
    execute_with_retry,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class KernelEventHandler:
    _valkey_stream: ValkeyStreamClient
    _registry: AgentRegistry
    _db: ExtendedAsyncSAEngine

    def __init__(
        self, valkey_stream: ValkeyStreamClient, registry: AgentRegistry, db: ExtendedAsyncSAEngine
    ) -> None:
        self._valkey_stream = valkey_stream
        self._registry = registry
        self._db = db

    async def handle_kernel_log(
        self,
        context: None,
        source: AgentId,
        event: DoSyncKernelLogsEvent,
    ) -> None:
        # The log data is at most 10 MiB.
        log_buffer = BytesIO()
        try:
            list_size = await self._valkey_stream.container_log_len(container_id=event.container_id)
            if list_size is None:
                # The log data is expired due to a very slow event delivery.
                # (should never happen!)
                log.warning(
                    "tried to store console logs for cid:{}, but the data is expired",
                    event.container_id,
                )
                return
            for _ in range(list_size):
                # Read chunk-by-chunk to allow interleaving with other Redis operations.
                chunks = await self._valkey_stream.pop_container_logs(
                    container_id=event.container_id
                )
                if chunks is None:  # maybe missing
                    log_buffer.write(b"(container log unavailable)\n")
                    break
                for chunk in chunks:
                    log_buffer.write(chunk)
            try:
                log_data = log_buffer.getvalue()

                async def _update_log() -> None:
                    async with self._db.begin() as conn:
                        update_query = (
                            sa.update(kernels)
                            .values(container_log=log_data)
                            .where(kernels.c.id == event.kernel_id)
                        )
                        await conn.execute(update_query)

                await execute_with_retry(_update_log)
            finally:
                # Clear the log data from Redis when done.
                await self._valkey_stream.clear_container_logs(container_id=event.container_id)
        finally:
            log_buffer.close()

    async def handle_kernel_creation_lifecycle(
        self,
        context: None,
        source: AgentId,
        event: (
            KernelPreparingAnycastEvent
            | KernelPullingAnycastEvent
            | KernelCreatingAnycastEvent
            | KernelStartedAnycastEvent
            | KernelCancelledAnycastEvent
        ),
    ) -> None:
        """
        Update the database and perform post_create_kernel() upon
        the events for each step of kernel creation.

        To avoid race condition between consumer and subscriber event handlers,
        we only have this handler to subscribe all kernel creation events,
        but distinguish which one to process using a unique creation_id
        generated when initiating the create_kernels() agent RPC call.
        """
        log.debug(
            "handle_kernel_creation_lifecycle: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )
        match event:
            case KernelPreparingAnycastEvent():
                # State transition is done by the DoPrepareEvent handler inside the scheduler-distpacher object.
                pass
            case KernelPullingAnycastEvent(kernel_id, session_id, reason=reason):
                async with self._db.connect() as db_conn:
                    await self._registry.mark_kernel_pulling(db_conn, kernel_id, session_id, reason)
            case KernelCreatingAnycastEvent(kernel_id, session_id, reason=reason):
                async with self._db.connect() as db_conn:
                    await self._registry.mark_kernel_creating(
                        db_conn, kernel_id, session_id, reason
                    )
            case KernelStartedAnycastEvent(
                kernel_id, session_id, reason=reason, creation_info=creation_info
            ):
                async with self._db.connect() as db_conn:
                    await self._registry.mark_kernel_running(
                        db_conn, kernel_id, session_id, reason, creation_info
                    )
            case KernelCancelledAnycastEvent():
                log.warning(f"Kernel cancelled, {event.reason = }")

    async def handle_kernel_preparing(
        self,
        context: None,
        source: AgentId,
        event: KernelPreparingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_preparing: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )
        # State transition is done by the DoPrepareEvent handler inside the scheduler-distpacher object.

    async def handle_kernel_pulling(
        self,
        context: None,
        source: AgentId,
        event: KernelPullingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_pulling: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )
        async with self._db.connect() as db_conn:
            await self._registry.mark_kernel_pulling(
                db_conn, event.kernel_id, event.session_id, event.reason
            )

    async def handle_kernel_creating(
        self,
        context: None,
        source: AgentId,
        event: KernelCreatingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_creating: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )
        async with self._db.connect() as db_conn:
            await self._registry.mark_kernel_creating(
                db_conn, event.kernel_id, event.session_id, event.reason
            )

    async def handle_kernel_started(
        self,
        context: None,
        source: AgentId,
        event: KernelStartedAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_started: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )
        async with self._db.connect() as db_conn:
            await self._registry.mark_kernel_running(
                db_conn, event.kernel_id, event.session_id, event.reason, event.creation_info
            )

    async def handle_kernel_cancelled(
        self,
        context: None,
        source: AgentId,
        event: KernelCancelledAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_cancelled: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

    async def handle_kernel_terminating(
        self,
        context: None,
        source: AgentId,
        event: KernelTerminatingAnycastEvent,
    ) -> None:
        # `destroy_kernel()` has already changed the kernel status to "TERMINATING".
        pass

    async def handle_kernel_terminated(
        self,
        context: None,
        source: AgentId,
        event: KernelTerminatedAnycastEvent,
    ) -> None:
        async with self._db.connect() as db_conn:
            await self._registry.mark_kernel_terminated(
                db_conn, event.kernel_id, event.session_id, event.reason, event.exit_code
            )

    async def handle_kernel_heartbeat(
        self,
        context: None,
        source: AgentId,
        event: KernelHeartbeatEvent,
    ) -> None:
        await self._registry.mark_kernel_heartbeat(event.kernel_id)



import logging
from collections import ChainMap
from datetime import datetime
from typing import Mapping, Optional, cast

import sqlalchemy as sa
from aiohttp import web
from sqlalchemy.engine.row import Row

from ai.backend.common.dto.manager.auth.field import AuthTokenType
from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.common.plugin.hook import ALL_COMPLETED, FIRST_COMPLETED, PASSED, HookPluginContext
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.unified import AuthConfig
from ai.backend.manager.data.auth.types import AuthorizationResult, SSHKeypair
from ai.backend.manager.errors.exceptions import (
    AuthorizationFailed,
    GenericBadRequest,
    GenericForbidden,
    InternalServerError,
    ObjectNotFound,
    PasswordExpired,
    RejectedByHook,
    UserNotFound,
)
from ai.backend.manager.models.group import association_groups_users, groups
from ai.backend.manager.models.keypair import (
    generate_keypair,
    generate_ssh_keypair,
    keypairs,
    validate_ssh_keypair,
)
from ai.backend.manager.models.user import (
    INACTIVE_USER_STATUSES,
    UserRole,
    UserRow,
    UserStatus,
    check_credential,
    compare_to_hashed_password,
    users,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_retry
from ai.backend.manager.services.auth.actions.authorize import (
    AuthorizeAction,
    AuthorizeActionResult,
)
from ai.backend.manager.services.auth.actions.generate_ssh_keypair import (
    GenerateSSHKeypairAction,
    GenerateSSHKeypairActionResult,
)
from ai.backend.manager.services.auth.actions.get_role import GetRoleAction, GetRoleActionResult
from ai.backend.manager.services.auth.actions.get_ssh_keypair import (
    GetSSHKeypairAction,
    GetSSHKeypairActionResult,
)
from ai.backend.manager.services.auth.actions.signout import SignoutAction, SignoutActionResult
from ai.backend.manager.services.auth.actions.signup import SignupAction, SignupActionResult
from ai.backend.manager.services.auth.actions.update_full_name import (
    UpdateFullNameAction,
    UpdateFullNameActionResult,
)
from ai.backend.manager.services.auth.actions.update_password import (
    UpdatePasswordAction,
    UpdatePasswordActionResult,
)
from ai.backend.manager.services.auth.actions.update_password_no_auth import (
    UpdatePasswordNoAuthAction,
    UpdatePasswordNoAuthActionResult,
)
from ai.backend.manager.services.auth.actions.upload_ssh_keypair import (
    UploadSSHKeypairAction,
    UploadSSHKeypairActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AuthService:
    _db: ExtendedAsyncSAEngine
    _hook_plugin_ctx: HookPluginContext

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        hook_plugin_ctx: HookPluginContext,
    ) -> None:
        self._db = db
        self._hook_plugin_ctx = hook_plugin_ctx

    async def get_role(self, action: GetRoleAction) -> GetRoleActionResult:
        group_role = None
        if action.group_id is not None:
            query = (
                # TODO: per-group role is not yet implemented.
                sa.select([association_groups_users.c.group_id])
                .select_from(association_groups_users)
                .where(
                    (association_groups_users.c.group_id == action.group_id)
                    & (association_groups_users.c.user_id == action.user_id),
                )
            )
            async with self._db.begin() as conn:
                result = await conn.execute(query)
                row = result.first()
                if row is None:
                    raise ObjectNotFound(
                        extra_msg="No such project or you are not the member of it.",
                        object_name="project (user group)",
                    )
            group_role = "user"

        return GetRoleActionResult(
            global_role="superadmin" if action.is_superadmin else "user",
            domain_role="admin" if action.is_admin else "user",
            group_role=group_role,
        )

    async def authorize(self, action: AuthorizeAction) -> AuthorizeActionResult:
        if action.type != AuthTokenType.KEYPAIR:
            # other types are not implemented yet.
            raise InvalidAPIParameters("Unsupported authorization type")

        params = action.hook_params
        hook_result = await self._hook_plugin_ctx.dispatch(
            "AUTHORIZE",
            (action.request, params),
            return_when=FIRST_COMPLETED,
        )
        if hook_result.status != PASSED:
            raise RejectedByHook.from_hook_result(hook_result)
        elif hook_result.result:
            # Passed one of AUTHORIZED hook
            user = hook_result.result
        else:
            # No AUTHORIZE hook is defined (proceed with normal login)
            user = await check_credential(
                db=self._db,
                domain=action.domain_name,
                email=action.email,
                password=action.password,
            )
        if user is None:
            raise AuthorizationFailed("User credential mismatch.")
        if user["status"] == UserStatus.BEFORE_VERIFICATION:
            raise AuthorizationFailed("This account needs email verification.")
        if user["status"] in INACTIVE_USER_STATUSES:
            raise AuthorizationFailed("User credential mismatch.")
        await self._check_password_age(user, action.auth_config)
        async with self._db.begin_session() as db_session:
            user_row = await UserRow.query_user_by_uuid(user["uuid"], db_session)
            if user_row is None:
                raise UserNotFound(extra_data=user["uuid"])
            main_keypair_row = user_row.get_main_keypair_row()
        if main_keypair_row is None:
            raise AuthorizationFailed("No API keypairs found.")
        # [Hooking point for POST_AUTHORIZE]
        # The hook handlers should accept a tuple of the request, user, and keypair objects.
        hook_result = await self._hook_plugin_ctx.dispatch(
            "POST_AUTHORIZE",
            (action.request, params, user, main_keypair_row.mapping),
            return_when=FIRST_COMPLETED,
        )
        if hook_result.status != PASSED:
            raise RejectedByHook.from_hook_result(hook_result)
        if hook_result.result is not None and isinstance(hook_result.result, web.StreamResponse):
            return AuthorizeActionResult(
                stream_response=hook_result.result,
                authorization_result=None,
            )

        return AuthorizeActionResult(
            stream_response=None,
            authorization_result=AuthorizationResult(
                access_key=main_keypair_row.access_key,
                secret_key=main_keypair_row.secret_key,
                user_id=user["uuid"],
                role=user["role"],
                status=user["status"],
            ),
        )

    async def signup(self, action: SignupAction) -> SignupActionResult:
        params = action.hook_params
        hook_result = await self._hook_plugin_ctx.dispatch(
            "PRE_SIGNUP",
            (params,),
            return_when=ALL_COMPLETED,
        )
        if hook_result.status != PASSED:
            raise RejectedByHook.from_hook_result(hook_result)
        else:
            # Merge the hook results as a single map.
            user_data_overriden = ChainMap(*cast(Mapping, hook_result.result))

        # [Hooking point for VERIFY_PASSWORD_FORMAT with the ALL_COMPLETED requirement]
        # The hook handlers should accept the request and whole ``params` dict.
        # They should return None if the validation is successful and raise the
        # Reject error otherwise.
        hook_result = await self._hook_plugin_ctx.dispatch(
            "VERIFY_PASSWORD_FORMAT",
            (action.request, params),
            return_when=ALL_COMPLETED,
        )
        if hook_result.status != PASSED:
            hook_result.reason = hook_result.reason or "invalid password format"
            raise RejectedByHook.from_hook_result(hook_result)

        async with self._db.begin() as conn:
            # Check if email already exists.
            query = sa.select([users]).select_from(users).where((users.c.email == action.email))
            result = await conn.execute(query)
            row = result.first()
            if row is not None:
                raise GenericBadRequest("Email already exists")

            # Create a user.
            data = {
                "domain_name": action.domain_name,
                "username": action.username if action.username is not None else action.email,
                "email": action.email,
                "password": action.password,
                "need_password_change": False,
                "full_name": action.full_name if action.full_name is not None else "",
                "description": action.description if action.description is not None else "",
                "status": UserStatus.ACTIVE,
                "status_info": "user-signup",
                "role": UserRole.USER,
                "integration_id": None,
                "resource_policy": "default",
                "sudo_session_enabled": False,
            }
            if user_data_overriden:
                for key, val in user_data_overriden.items():
                    if (
                        key in data  # take only valid fields
                        and key != "resource_policy"  # resource_policy in user_data is for keypair
                    ):
                        data[key] = val
            query = users.insert().values(data)
            result = await conn.execute(query)
            if result.rowcount > 0:
                checkq = users.select().where(users.c.email == action.email)
                result = await conn.execute(checkq)
                user = result.first()
                # Create user's first access_key and secret_key.
                ak, sk = generate_keypair()
                resource_policy = user_data_overriden.get("resource_policy", "default")
                kp_data = {
                    "user_id": action.email,
                    "access_key": ak,
                    "secret_key": sk,
                    "is_active": True if data.get("status") == UserStatus.ACTIVE else False,
                    "is_admin": False,
                    "resource_policy": resource_policy,
                    "rate_limit": 1000,
                    "num_queries": 0,
                    "user": user.uuid,
                }
                query = keypairs.insert().values(kp_data)
                await conn.execute(query)

                # Add user to the default group.
                group_name = user_data_overriden.get("group", "default")
                query = (
                    sa.select([groups.c.id])
                    .select_from(groups)
                    .where(groups.c.domain_name == action.domain_name)
                    .where(groups.c.name == group_name)
                )
                result = await conn.execute(query)
                grp = result.first()
                if grp is not None:
                    values = [{"user_id": user.uuid, "group_id": grp.id}]
                    query = association_groups_users.insert().values(values)
                    await conn.execute(query)
            else:
                raise InternalServerError("Error creating user account")

        # [Hooking point for POST_SIGNUP as one-way notification]
        # The hook handlers should accept a tuple of the user email,
        # the new user's UUID, and a dict with initial user's preferences.
        initial_user_prefs = {
            "lang": action.request.headers.get("Accept-Language", "en-us").split(",")[0].lower(),
        }
        await self._hook_plugin_ctx.notify(
            "POST_SIGNUP",
            (action.email, user.uuid, initial_user_prefs),
        )
        return SignupActionResult(
            user_id=user.uuid,
            access_key=ak,
            secret_key=sk,
        )

    async def signout(self, action: SignoutAction) -> SignoutActionResult:
        if action.email != action.requester_email:
            raise GenericForbidden("Not the account owner")
        email = action.email
        result = await check_credential(
            db=self._db,
            domain=action.domain_name,
            email=email,
            password=action.password,
        )
        if result is None:
            raise GenericBadRequest("Invalid email and/or password")
        async with self._db.begin() as conn:
            # Inactivate the user.
            query = users.update().values(status=UserStatus.INACTIVE).where(users.c.email == email)
            await conn.execute(query)
            # Inactivate every keypairs of the user.
            query = keypairs.update().values(is_active=False).where(keypairs.c.user_id == email)
            await conn.execute(query)

        return SignoutActionResult(success=True)

    async def update_full_name(self, action: UpdateFullNameAction) -> UpdateFullNameActionResult:
        email = action.email
        domain_name = action.domain_name
        async with self._db.begin() as conn:
            query = (
                sa.select([users])
                .select_from(users)
                .where(
                    (users.c.email == email) & (users.c.domain_name == domain_name),
                )
            )
            result = await conn.execute(query)
            user = result.first()
            if user is None:
                return UpdateFullNameActionResult(success=False)

            # If user is not null, then it updates user full_name.
            data = {
                "full_name": action.full_name,
            }
            update_query = users.update().values(data).where(users.c.email == email)
            await conn.execute(update_query)
            return UpdateFullNameActionResult(success=True)

    async def update_password(self, action: UpdatePasswordAction) -> UpdatePasswordActionResult:
        domain_name = action.domain_name
        email = action.email
        log_fmt = "AUTH.UPDATE_PASSWORD(d:{}, email:{})"
        log_args = (domain_name, email)
        user = await check_credential(
            db=self._db,
            domain=domain_name,
            email=email,
            password=action.old_password,
        )
        if user is None:
            log.info(log_fmt + ": old password mismtach", *log_args)
            raise AuthorizationFailed("Old password mismatch")
        if action.new_password != action.new_password_confirm:
            log.info(log_fmt + ": new password mismtach", *log_args)
            return UpdatePasswordActionResult(
                success=False,
                message="new password mismatch",
            )

        # [Hooking point for VERIFY_PASSWORD_FORMAT with the ALL_COMPLETED requirement]
        # The hook handlers should accept the request and whole ``params` dict.
        # They should return None if the validation is successful and raise the
        # Reject error otherwise.
        hook_result = await self._hook_plugin_ctx.dispatch(
            "VERIFY_PASSWORD_FORMAT",
            (action.request, action.hook_params),
            return_when=ALL_COMPLETED,
        )
        if hook_result.status != PASSED:
            hook_result.reason = hook_result.reason or "invalid password format"
            raise RejectedByHook.from_hook_result(hook_result)

        async with self._db.begin() as conn:
            # Update user password.
            data = {
                "password": action.new_password,
                "need_password_change": False,
                "password_changed_at": sa.func.now(),
            }
            query = users.update().values(data).where(users.c.email == email)
            await conn.execute(query)

        return UpdatePasswordActionResult(
            success=True,
            message="Password updated successfully",
        )

    async def update_password_no_auth(
        self, action: UpdatePasswordNoAuthAction
    ) -> UpdatePasswordNoAuthActionResult:
        if action.auth_config is None or action.auth_config.max_password_age is None:
            raise GenericBadRequest("Unsupported function.")

        checked_user = await check_credential(
            db=self._db,
            domain=action.domain_name,
            email=action.email,
            password=action.current_password,
        )
        if checked_user is None:
            raise AuthorizationFailed("User credential mismatch.")
        new_password = action.new_password
        if compare_to_hashed_password(new_password, checked_user["password"]):
            raise AuthorizationFailed("Cannot update to the same password as an existing password.")

        # [Hooking point for VERIFY_PASSWORD_FORMAT with the ALL_COMPLETED requirement]
        # The hook handlers should accept the request and whole ``params` dict.
        # They should return None if the validation is successful and raise the
        # Reject error otherwise.
        hook_result = await self._hook_plugin_ctx.dispatch(
            "VERIFY_PASSWORD_FORMAT",
            (action.request, action.hook_params),
            return_when=ALL_COMPLETED,
        )
        if hook_result.status != PASSED:
            hook_result.reason = hook_result.reason or "invalid password format"
            raise RejectedByHook.from_hook_result(hook_result)

        async def _update() -> datetime:
            async with self._db.begin() as conn:
                # Update user password.
                data = {
                    "password": new_password,
                    "need_password_change": False,
                    "password_changed_at": sa.func.now(),
                }
                query = (
                    sa.update(users)
                    .values(data)
                    .where(users.c.uuid == checked_user["uuid"])
                    .returning(users.c.password_changed_at)
                )
                result = await conn.execute(query)
                return result.scalar()

        changed_at = await execute_with_retry(_update)

        return UpdatePasswordNoAuthActionResult(
            user_id=checked_user["uuid"],
            password_changed_at=changed_at,
        )

    async def get_ssh_keypair(self, action: GetSSHKeypairAction) -> GetSSHKeypairActionResult:
        async with self._db.begin() as conn:
            # Get SSH public key. Return partial string from the public key just for checking.
            query = sa.select([keypairs.c.ssh_public_key]).where(
                keypairs.c.access_key == action.access_key
            )
            pubkey = await conn.scalar(query)

            return GetSSHKeypairActionResult(public_key=pubkey)

    async def generate_ssh_keypair(
        self, action: GenerateSSHKeypairAction
    ) -> GenerateSSHKeypairActionResult:
        async with self._db.begin() as conn:
            pubkey, privkey = generate_ssh_keypair()
            data = {
                "ssh_public_key": pubkey,
                "ssh_private_key": privkey,
            }
            query = keypairs.update().values(data).where(keypairs.c.access_key == action.access_key)
            await conn.execute(query)

        return GenerateSSHKeypairActionResult(
            ssh_keypair=SSHKeypair(
                ssh_public_key=pubkey,
                ssh_private_key=privkey,
            )
        )

    async def upload_ssh_keypair(
        self, action: UploadSSHKeypairAction
    ) -> UploadSSHKeypairActionResult:
        privkey = action.private_key
        pubkey = action.public_key
        is_valid, err_msg = validate_ssh_keypair(privkey, pubkey)
        if not is_valid:
            raise InvalidAPIParameters(err_msg)
        async with self._db.begin() as conn:
            data = {
                "ssh_public_key": pubkey,
                "ssh_private_key": privkey,
            }
            query = keypairs.update().values(data).where(keypairs.c.access_key == action.access_key)
            await conn.execute(query)

        return UploadSSHKeypairActionResult(
            ssh_keypair=SSHKeypair(
                ssh_public_key=pubkey,
                ssh_private_key=privkey,
            ),
        )

    async def _check_password_age(self, user: Row, auth_config: Optional[AuthConfig]) -> None:
        if (
            auth_config is not None
            and (max_password_age := auth_config.max_password_age) is not None
        ):
            password_changed_at: datetime = user.users_password_changed_at

            async with self._db.begin_readonly() as db_conn:
                current_dt: datetime = await db_conn.scalar(sa.select(sa.func.now()))
                if password_changed_at + max_password_age < current_dt:
                    # Force user to update password
                    raise PasswordExpired(
                        extra_msg=f"Password expired on {password_changed_at + max_password_age}."
                    )



from __future__ import annotations

from typing import TYPE_CHECKING

import attrs

from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.message_queue.queue import AbstractMessageQueue
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.plugin.event import EventDispatcherPluginContext
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscovery,
    ServiceDiscoveryLoop,
)
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.plugin.network import NetworkPluginContext
from ai.backend.manager.scheduler.dispatcher import SchedulerDispatcher
from ai.backend.manager.service.base import ServicesContext
from ai.backend.manager.services.processors import Processors

if TYPE_CHECKING:
    from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
    from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
    from ai.backend.common.plugin.hook import HookPluginContext
    from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
    from ai.backend.common.types import RedisConnectionInfo

    from ..agent_cache import AgentRPCCache
    from ..idle import IdleCheckerHost
    from ..models.storage import StorageSessionManager
    from ..models.utils import ExtendedAsyncSAEngine
    from ..plugin.webapp import WebappPluginContext
    from ..registry import AgentRegistry
    from ..types import DistributedLockFactory
    from .types import CORSOptions


class BaseContext:
    pass


@attrs.define(slots=True, auto_attribs=True, init=False)
class RootContext(BaseContext):
    pidx: int
    db: ExtendedAsyncSAEngine
    distributed_lock_factory: DistributedLockFactory
    event_dispatcher: EventDispatcher
    event_fetcher: EventFetcher
    event_producer: EventProducer
    etcd: AsyncEtcd
    redis_live: RedisConnectionInfo
    redis_stat: RedisConnectionInfo
    redis_image: RedisConnectionInfo
    redis_stream: RedisConnectionInfo
    valkey_stream: ValkeyStreamClient
    redis_lock: RedisConnectionInfo
    config_provider: ManagerConfigProvider
    cors_options: CORSOptions

    webapp_plugin_ctx: WebappPluginContext
    idle_checker_host: IdleCheckerHost
    storage_manager: StorageSessionManager
    hook_plugin_ctx: HookPluginContext
    network_plugin_ctx: NetworkPluginContext
    event_dispatcher_plugin_ctx: EventDispatcherPluginContext
    services_ctx: ServicesContext

    registry: AgentRegistry
    agent_cache: AgentRPCCache
    scheduler_dispatcher: SchedulerDispatcher

    error_monitor: ErrorPluginContext
    stats_monitor: StatsPluginContext
    background_task_manager: BackgroundTaskManager
    metrics: CommonMetricRegistry
    processors: Processors
    event_hub: EventHub
    message_queue: AbstractMessageQueue
    service_discovery: ServiceDiscovery
    sd_loop: ServiceDiscoveryLoop

    def __init__(self, *, metrics: CommonMetricRegistry = CommonMetricRegistry(), **kwargs) -> None:
        super().__init__(**kwargs)
        self.metrics = metrics



import uuid
from dataclasses import dataclass


@dataclass
class SSHKeypair:
    ssh_public_key: str
    ssh_private_key: str


@dataclass
class AuthorizationResult:
    user_id: uuid.UUID
    access_key: str
    secret_key: str
    role: str
    status: str



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.common.types import AgentId
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.agent.actions.base import AgentAction


@dataclass
class WatcherAgentStartAction(AgentAction):
    agent_id: AgentId

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.agent_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "watcher_agent_start"


@dataclass
class WatcherAgentStartActionResult(BaseActionResult):
    # TODO: Add proper type
    resp: Any
    agent_id: AgentId

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.agent_id)



import logging

from ai.backend.common.events.event_types.session.anycast import (
    DoRecalculateUsageEvent,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.registry import AgentRegistry

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class UsageEventHandler:
    _agent_registry: AgentRegistry

    def __init__(self, agent_registry: AgentRegistry) -> None:
        self._agent_registry = agent_registry

    async def handle_do_recalculate_usage(
        self, context: None, agent_id: str, ev: DoRecalculateUsageEvent
    ) -> None:
        await self._agent_registry.recalc_resource_usage(do_fullscan=True)



import logging

from ai.backend.common.events.event_types.agent.anycast import AgentStartedEvent
from ai.backend.common.events.event_types.schedule.anycast import (
    DoCheckPrecondEvent,
    DoScaleEvent,
    DoScheduleEvent,
    DoStartSessionEvent,
)
from ai.backend.common.events.event_types.session.anycast import (
    DoUpdateSessionStatusEvent,
    SessionEnqueuedAnycastEvent,
    SessionTerminatedAnycastEvent,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.scheduler.dispatcher import SchedulerDispatcher

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ScheduleEventHandler:
    _scheduler_dispatcher: SchedulerDispatcher

    def __init__(self, scheduler_dispatcher: SchedulerDispatcher) -> None:
        self._scheduler_dispatcher = scheduler_dispatcher

    async def handle_session_enqueued(
        self, context: None, agent_id: str, ev: SessionEnqueuedAnycastEvent
    ) -> None:
        await self._scheduler_dispatcher.schedule(ev.event_name())

    async def handle_session_terminated(
        self, context: None, agent_id: str, ev: SessionTerminatedAnycastEvent
    ) -> None:
        await self._scheduler_dispatcher.schedule(ev.event_name())

    async def handle_agent_started(
        self, context: None, agent_id: str, ev: AgentStartedEvent
    ) -> None:
        await self._scheduler_dispatcher.schedule(ev.event_name())

    async def handle_do_schedule(self, context: None, agent_id: str, ev: DoScheduleEvent) -> None:
        await self._scheduler_dispatcher.schedule(ev.event_name())

    async def handle_do_start_session(
        self, context: None, agent_id: str, ev: DoStartSessionEvent
    ) -> None:
        await self._scheduler_dispatcher.start(ev.event_name())

    async def handle_do_check_precond(
        self, context: None, agent_id: str, ev: DoCheckPrecondEvent
    ) -> None:
        await self._scheduler_dispatcher.check_precond(ev.event_name())

    async def handle_do_scale(self, context: None, agent_id: str, ev: DoScaleEvent) -> None:
        await self._scheduler_dispatcher.scale_services(ev.event_name())

    async def handle_do_update_session_status(
        self, context: None, agent_id: str, ev: DoUpdateSessionStatusEvent
    ) -> None:
        await self._scheduler_dispatcher.update_session_status()
